import time

from typing import List

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class DummyTask(Task):
    """
    This is an example task that you can copy to implement your own tasks. 
    The following methods are available from the base Task class:

        - self.parameter(name)
        - parameterValuesAsString()
        - self.dataManager()
        - self.addInfo(message)
        - self.addError(message)
        - self.addWarning(message)
        - self.updateProgress(step, nrSteps)
        - self.readFromCache(file))
        - self.writeToCache(file, fileObject)
        - self.generateTimestampForFileSetName(name)
    """
    def __init__(self) -> None:
        super(DummyTask, self).__init__()

    def execute(self) -> None:

        # Get parameters needed for this task
        nrIterations = self.parameter('nrIterations').value()
        
        # Do iterations of the task
        for i in range(nrIterations):
        
            # Check if task was canceled first
            if self.statusIsCanceled():
                self.addInfo('Canceling task...')
                break

            # Do your processing here...
            self.addInfo(f'Processing iteration {i}/{nrIterations}')

            # Update progress based on nr. steps required. This will automatically
            # send sigals/events to the task widget
            self.updateProgress(step=i, nrSteps=nrIterations)

            # If necessary wait a bit
            time.sleep(1)


from typing import Any

from PySide6.QtWidgets import QWidget, QLabel, QSizePolicy
from PySide6.QtGui import QFont

from mosamaticdesktop.tasks.parameter import Parameter


class DescriptionParameter(Parameter):
    def __init__(self, name: str, description: str, parent: QWidget=None) -> None:
        super(DescriptionParameter, self).__init__(
            name=name, labelText=description, optional=True, parent=parent)
        self.initUi()

    def initUi(self) -> None:
        font = QFont()
        label = QLabel(self.labelText())
        label.setFont(font)
        label.setWordWrap(True)
        policy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        label.setSizePolicy(policy)
        self.layout().addWidget(label)

    def copy(self):
        return DescriptionParameter(
            name=self.name(), 
            description=self.labelText()
        )


from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.dummytask.dummytask import DummyTask


class DummyTaskWidget(TaskWidget):
    def __init__(self, progressBar: QProgressBar) -> None:
        super(DummyTaskWidget, self).__init__(taskType=DummyTask, progressBar=progressBar)
        self.addDescriptionParameter(
            name='description',
            description='This is some description of the dummy task'
        )
        self.addIntegerParameter(
            name='nrIterations', 
            labelText='Nr. Iterations',
            optional=False,
            visible=True,
            defaultValue=10,
            minimum=0,
            maximum=100,
            step=1,
        )
    
    def validate(self) -> None:
        # self.showValidationError(parameterName='Some parameter', message='Something wrong')
        pass


import os
import zipfile

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.utils import createNameWithTimestamp
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class CreateArchiveTask(Task):
    def __init__(self) -> None:
        super(CreateArchiveTask, self).__init__()

    def execute(self) -> None:

        # Get parameters needed for this task
        inputFileSetName = self.parameter(name='inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=False)

        zipFileName = createNameWithTimestamp(inputFileSet.name()) + '.zip'
        outputZipFilePath = os.path.join(outputFileSetPath, zipFileName)

        step = 0
        files = inputFileSet.files()
        nrSteps = len(files)
        with zipfile.ZipFile(outputZipFilePath, 'w') as zipObj:
            for file in files:
                zipObj.write(file.path(), arcname=os.path.basename(file.path()))

                # Update progress
                self.updateProgress(step=step, nrSteps=nrSteps)
                step += 1

        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        self.addInfo('Finished')



from typing import List

from PySide6.QtWidgets import QWidget, QDialog, QVBoxLayout, QListWidget, QPushButton, QTabWidget


class TaskRunInfoDialog(QDialog):
    def __init__(self, taskName: str, errors: List[str], warnings: List[str], info: List[str], parent: QWidget=None) -> None:
        super(TaskRunInfoDialog, self).__init__(parent)
        self._taskName = taskName
        self._errors = errors
        self._warnings = warnings
        self._info = info
        self._tabWidget = None
        self._errorListWidget = None
        self._warningsListWidget = None
        self._infoListWidget = None
        self.initUi()
        self.highlightTab()

    def initUi(self) -> None:
        self._errorListWidget = QListWidget(self)
        for error in self._errors:
            self._errorListWidget.addItem(error)
        self._warningsListWidget = QListWidget(self)
        for warning in self._warnings:
            self._warningsListWidget.addItem(warning)
        self._infoListWidget = QListWidget(self)
        for info in self._info:
            self._infoListWidget.addItem(info)
        self._tabWidget = QTabWidget(self)
        self._tabWidget.addTab(self._errorListWidget, 'Errors')
        self._tabWidget.addTab(self._warningsListWidget, 'Warnings')
        self._tabWidget.addTab(self._infoListWidget, 'Info')
        self._tabWidget.setFixedWidth(600)
        closeButton = QPushButton('Close', self)
        closeButton.clicked.connect(self.close)
        layout = QVBoxLayout()
        layout.addWidget(self._tabWidget)
        layout.addWidget(closeButton)
        self.setLayout(layout)
        self.setWindowTitle(f'Run Info for Task {self._taskName}')

    def highlightTab(self) -> None:
        if len(self._errors) > 0:
            self._tabWidget.setCurrentIndex(0)
        elif len(self._warnings) > 0:
            self._tabWidget.setCurrentIndex(1)
        else:
            self._tabWidget.setCurrentIndex(2)

    def show(self):
        return self.exec()
    
    def close(self):
        self.accept()


from typing import Any, List

from PySide6.QtWidgets import QWidget, QLabel, QComboBox

from mosamaticdesktop.tasks.parameter import Parameter


class OptionGroupParameter(Parameter):
    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, options: List[str]=[], parent: QWidget=None) -> None:
        super(OptionGroupParameter, self).__init__(
            name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, parent=parent)
        self._options = options
        self._optionGroupComboBox = None
        self.initUi()

    def options(self) -> List[str]:
        return self._options

    def initUi(self) -> None:
        self._optionGroupComboBox = QComboBox(self)
        self._optionGroupComboBox.addItem(None)
        for option in self._options:
            self._optionGroupComboBox.addItem(option)
        if self.defaultValue() is not None:
            self._optionGroupComboBox.setCurrentText(self.defaultValue())
            self.setValue(self.defaultValue())
        self._optionGroupComboBox.currentTextChanged.connect(self.currentTextChanged)
        self.layout().addWidget(QLabel(self.labelText()))
        self.layout().addWidget(self._optionGroupComboBox)

    def currentTextChanged(self, text: str) -> None:
        self.setValue(text)

    def copy(self):
        return OptionGroupParameter(
            name=self.name(), 
            labelText=self.labelText(), 
            optional=self.optional(), 
            visible=self.visible(), 
            defaultValue=self.defaultValue(),
            options=self.options(),
        )


import os
import json
import shutil
import zipfile
import pydicom
import pydicom.errors
import numpy as np

from typing import List, Any

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.ai.tensorflowmodel import TensorFlowModel
from mosamaticdesktop.data.file import File
from mosamaticdesktop.logger import Logger
from mosamaticdesktop.utils import getPixelsFromDicomObject, convertLabelsTo157
from mosamaticdesktop.utils import normalizeBetween, Configuration
from mosamaticdesktop.utils import readFromCache, writeToCache

LOGGER = Logger()


class MuscleFatSegmentationTask(Task):
    ARGMAX = 0
    PROBABILITIES = 1

    def __init__(self) -> None:
        super(MuscleFatSegmentationTask, self).__init__()        

    def loadModelFiles(self, files: List[File]) -> List[Any]:
        tfLoaded = False
        configuration = Configuration()
        for file in files:
            if file.name() == 'model.zip':
                content = readFromCache(file=file)
                if not content:
                    if not tfLoaded:
                        import tensorflow as tf # Only load TensorFlow package if necessary (takes some time)
                        tfLoaded = True
                    tensorFlowModelFileDirectory = configuration.taskConfigSubDirectory(taskName=__class__.__name__, dirName='tensorFlowModelFiles')
                    with zipfile.ZipFile(file.path()) as zipObj:
                        zipObj.extractall(path=tensorFlowModelFileDirectory)
                    tensorFlowModel = TensorFlowModel()
                    tensorFlowModel.load(modelFilePath=tensorFlowModelFileDirectory)
                    content = writeToCache(file=file, fileObject=tensorFlowModel)
                tensorFlowModel = content.fileObject()
            elif file.name() == 'contour_model.zip':
                content = readFromCache(file=file)
                if not content:
                    if not tfLoaded:
                        import tensorflow as tf # Only load TensorFlow package if necessary (takes some time)
                        tfLoaded = True
                    tensorFlowModelFileDirectory = configuration.taskConfigSubDirectory(taskName=__class__.__name__, dirName='tensorFlowModelFiles')
                    with zipfile.ZipFile(file.path()) as zipObj:
                        zipObj.extractall(path=tensorFlowModelFileDirectory)
                    tensorFlowContourModel = TensorFlowModel()
                    tensorFlowContourModel.load(modelFilePath=tensorFlowModelFileDirectory)
                    content = writeToCache(file=file, fileObject=tensorFlowContourModel)
                tensorFlowContourModel = content.fileObject()
            elif file.name() == 'params.json':
                content = readFromCache(file=file)
                if not content:
                    with open(file.path(), 'r') as f:
                        parameters = json.load(f)
                        content = writeToCache(file=file, fileObject=parameters)
                parameters = content.fileObject()
            else:
                pass
        return [tensorFlowModel, tensorFlowContourModel, parameters]

    def predictContour(self, contourModel, sourceImage, parameters):
        ct = np.copy(sourceImage)
        ct = normalizeBetween(ct, parameters['min_bound_contour'], parameters['max_bound_contour'])
        img2 = np.expand_dims(ct, 0)
        img2 = np.expand_dims(img2, -1)
        pred = contourModel.predict([img2])
        predSqueeze = np.squeeze(pred)
        pred_max = predSqueeze.argmax(axis=-1)
        mask = np.uint8(pred_max)
        return mask
    
    def execute(self) -> None:

        # Get input fileset
        inputFileSetName = self.parameter('inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
        tensorFlowModelFileSetName = self.parameter('tensorFlowModelFileSetName').value()
        tensorFlowModelFileSet = self.dataManager().fileSetByName(tensorFlowModelFileSetName)
        # Setup output fileset path
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)

        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        self.addInfo(f'Overwrite output fileset: {overwriteOutputFileSet}')
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=False)
        self.addInfo(f'Output fileset path: {outputFileSetPath}')

        # Get mode (ARGMAX or PROBABILITIES)
        modeText = self.parameter('mode').value()
        mode = MuscleFatSegmentationTask.ARGMAX if modeText == 'ARGMAX' else MuscleFatSegmentationTask.PROBABILITIES
        self.addInfo(f'Mode: {modeText}')

        # Load TensorFlow model files
        model, contourModel, parameters = self.loadModelFiles(files=tensorFlowModelFileSet.files())
        if model and parameters:

            # Start iterating of the files
            step = 0
            files = inputFileSet.files()
            segmentationFiles = []
            nrSteps = len(files)
            for file in files:

                # Check if task was canceled first
                if self.statusIsCanceled():
                    self.addInfo('Canceling task...')
                    break

                try:
                    # Read DICOM file (from cache first) and decompress if needed
                    content = readFromCache(file=file)
                    if not content:
                        p = pydicom.dcmread(file.path())
                        p.decompress()
                        content = writeToCache(file, p)
                    p = content.fileObject()

                    # Get pixels from DICOM file and normalize to positive range
                    img1 = getPixelsFromDicomObject(p, normalize=True)

                    # If contour model provided, apply it to detect abdominal contour
                    if contourModel:
                        mask = self.predictContour(contourModel=contourModel, sourceImage=img1, parameters=parameters)
                        img1 = normalizeBetween(img=img1, minBound=parameters['min_bound'], maxBound=parameters['max_bound'])
                        img1 = img1 * mask
                    else:
                        img1 = normalizeBetween(img=img1, minBound=parameters['min_bound'], maxBound=parameters['max_bound'])

                    img1 = img1.astype(np.float32)
                    img2 = np.expand_dims(img1, 0)
                    img2 = np.expand_dims(img2, -1)
                    pred = model.predict([img2]) ##### Move to separate AI class!
                    predSqueeze = np.squeeze(pred)

                    # Generate predicted output. Can be ARGMAX (pixel value with maximum probability) or
                    # PROBABILITIES, i.e, the individual class probabilities (muscle, SAT and VAT) in 
                    # each pixel
                    if mode == MuscleFatSegmentationTask.ARGMAX:
                        predMax = predSqueeze.argmax(axis=-1)
                        predMax = convertLabelsTo157(labelImage=predMax)
                        segmentationFile = os.path.join(outputFileSetPath, f'{file.name()}.seg.npy')
                        segmentationFiles.append(segmentationFile)
                        np.save(segmentationFile, predMax)
                    elif mode == MuscleFatSegmentationTask.PROBABILITIES:
                        segmentationFile = os.path.join(outputFileSetPath, f'{file.name()}.seg.prob.npy')
                        segmentationFiles.append(segmentationFile)
                        np.save(segmentationFile, predSqueeze)

                except pydicom.errors.InvalidDicomError:
                    self.addWarning(f'Skipping non-DICOM: {file.path()}')

                # Update progress for this iteration         
                self.updateProgress(step=step, nrSteps=nrSteps)
                step += 1

        # Build output fileset
        outputFileSet = self.dataManager().createFileSet(fileSetPath=outputFileSetPath)

        # Cache all files in the output fileset because it's very likely we'll be using these
        # files again for further processing and visualization
        for file in outputFileSet.files():
            if file.name().endswith('.seg.npy') or file.name().endswith('.seg.prob.npy'):
                writeToCache(file=file, fileObject=np.load(file.path()))

        self.addInfo('Finished')


from typing import Any

from PySide6.QtWidgets import QWidget, QLabel, QComboBox

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.data.datamanager import DataManager


class FileSetParameter(Parameter):
    class ComboBox(QComboBox):
        def __init__(self, parent: QWidget=None) -> None:
            super(FileSetParameter.ComboBox, self).__init__(parent=parent)
            self._dataManager = DataManager()
            self.loadItems()

        def loadItems(self) -> None:
            self.addItem(None)
            fileSets = self._dataManager.fileSets()
            for fileSet in fileSets:
                self.addItem(fileSet.name())

        def showPopup(self) -> None:
            self.clear()
            self.loadItems()
            return super().showPopup()

    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, parent: QWidget=None) -> None:
        super(FileSetParameter, self).__init__(
            name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, parent=parent)
        if self.defaultValue() is not None:
            self.setValue(self.defaultValue())
        self.initUi()

    def initUi(self) -> None:
        label = QLabel(self.labelText())
        comboBox = FileSetParameter.ComboBox(self)
        comboBox.currentTextChanged.connect(self.currentTextChanged)
        self.layout().addWidget(label)
        self.layout().addWidget(comboBox)

    def currentTextChanged(self, text: str) -> None:
        self.setValue(text)

    def copy(self):
        return FileSetParameter(
            name=self.name(), 
            labelText=self.labelText(), 
            optional=self.optional(), 
            visible=self.visible(), 
            defaultValue=self.defaultValue()
        )


import threading

from typing import Dict, List, Any

from PySide6.QtCore import QObject, Signal

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.data.datamanager import DataManager
from mosamaticdesktop.logger import Logger
from mosamaticdesktop.utils import createNameWithTimestamp

LOGGER = Logger()


class Task:
    class TaskProgressSignal(QObject):
        progress = Signal(int)
        finished = Signal(bool)

    IDLE = 0
    START = 1
    RUNNING = 2
    CANCELING = 3
    CANCELED = 4
    FINISHED = 5
    ERROR = 6

    @classmethod
    def NAME(cls):
        # Returns class name of child classes
        return cls.__qualname__

    def __init__(self) -> None:
        self._status = Task.IDLE
        self._progress = 0
        self._thread = None
        self._errors = []
        self._warnings = []
        self._info = []
        self._parameters = None
        self._dataManager = DataManager()
        self._signal = self.TaskProgressSignal()

    def name(self) -> str:
        return self.__class__.__name__
    
    def logger(self) -> Logger:
        return LOGGER
    
    def setParameters(self, parameters: Dict[str, Parameter]) -> None:
        self._parameters = parameters
    
    def parameter(self, name: str) -> Parameter:
        if name in self._parameters.keys():
            return self._parameters[name]
        return None
    
    def parameterValuesAsString(self) -> str:
        string = ''
        for name in self._parameters.keys():
            value = self._parameters[name].value()
            string += f'{name}="{value}", '
        return string[:-2]
    
    def signal(self):
        return self._signal
    
    def dataManager(self) -> DataManager:
        return self._dataManager
    
    # Logging
    
    def errors(self) -> List[str]:
        return self._errors
    
    def addError(self, message: str, toStdOut: bool=True, cancel=False) -> None:
        if toStdOut:
            LOGGER.error(message)
        self._errors.append(message)
        if cancel:
            self.cancel()

    def hasErrors(self) -> bool:
        return len(self._errors) > 0
    
    def warnings(self) -> List[str]:
        return self._warnings
    
    def addWarning(self, message: str, toStdOut: bool=True) -> None:
        if toStdOut:
            LOGGER.warning(message)
        self._warnings.append(message)

    def hasWarnings(self) -> bool:
        return len(self._warnings) > 0
    
    def info(self) -> List[str]:
        return self._info
    
    def addInfo(self, message: str, toStdOut: bool=True) -> None:
        if toStdOut:
            LOGGER.info(message)
        self._info.append(message)

    def hasInfo(self) -> bool:
        return len(self._info) > 0
    
    # Status
    
    def status(self) -> int:
        return self._status
    
    def statusIsIdle(self) -> bool:
        return self._status == Task.IDLE
    
    def statusIsStart(self) -> bool:
        return self._status == Task.START
    
    def statusIsRunning(self) -> bool:
        return self._status == Task.RUNNING
    
    def statusIsCanceling(self) -> bool:
        return self._status == Task.CANCELING
    
    def statusIsCanceled(self) -> bool:
        return self._status == Task.CANCELED
    
    def statusIsFinished(self) -> bool:
        return self._status == Task.FINISHED
    
    def statusIsError(self) -> bool:
        return self._status == Task.ERROR
    
    def setStatus(self, status: int) -> None:
        self._status = status

    def setStatusIdle(self) -> None:
        self._status = Task.IDLE

    def setStatusStart(self) -> None:
        self._status = Task.START

    def setStatusRunning(self) -> None:
        self._status = Task.RUNNING

    def setStatusCanceling(self) -> None:
        self._status = Task.CANCELING

    def setStatusCanceled(self) -> None:
        self._status = Task.CANCELED

    def setStatusFinished(self) -> None:
        self._status = Task.FINISHED

    def setStatusError(self) -> None:
        self._status = Task.ERROR

    # Progress

    def progress(self) -> int:
        return self._progress
    
    def updateProgress(self, step: int, nrSteps: int) -> None:
        if self._signal:
            self._progress = int(((step + 1) / (nrSteps)) * 100)
            self._signal.progress.emit(self._progress)
            if self._progress >= 100:
                self._signal.finished.emit(True)
                self._signal = None

    # Execution
                
    def start(self) -> None:
        self.setStatusStart()
        self._thread = threading.Thread(target=self.run)
        self._thread.start()
        self.setStatusRunning()

    def run(self) -> None:
        self.addInfo(f'Executing {self.name()}({self.parameterValuesAsString()})...')
        self.execute()
        if not self.statusIsCanceled():
            if self.hasErrors():
                self.setStatusError()
            else:
                self.setStatusFinished()        

    def execute(self) -> None:
        raise NotImplementedError()    

    def cancel(self) -> None:
        self.setStatusCanceled()
        self._thread.join()

    # Miscellaneous

    def generateTimestampForFileSetName(self, name: str) -> str:
        return createNameWithTimestamp(prefix=f'{name}')


import os
import pytest

from mosamaticdesktop.data import engine

engine.DATABASE = 'db.test.sqlite3'
engine.ECHO = False


if __name__ == '__main__':
    if os.path.isfile(engine.DATABASE):
        os.remove(engine.DATABASE)
    pytest.main(['-s', 'src/app/test'])


import os

from PySide6.QtWidgets import QProgressBar

from typing import List

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.utils import ModuleLoader
from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


@singleton
class TaskWidgetManager:
    def __init__(self, progressBar: QProgressBar) -> None:
        self._progressBar = progressBar
        self._taskWidgetTypes = self.loadTaskWidgetTypes()
        self._taskWidgets = {}

    def taskNames(self) -> List[str]:
        return list(self._taskWidgetTypes.keys())
    
    def taskWidget(self, name) -> TaskWidget:
        if name in self._taskWidgetTypes.keys():
            if name not in self._taskWidgets.keys():
                self._taskWidgets[name] = self._taskWidgetTypes[name](progressBar=self._progressBar)
            return self._taskWidgets[name]
        LOGGER.error(f'TaskWidgetManager: task {name} does not exist')
        return None

    def loadTaskWidgetTypes(self):
        moduleDirectoryPath = os.path.dirname(os.path.realpath(__file__))
        LOGGER.info(f'TaskWidgetManager: loading tasks from {moduleDirectoryPath}...')
        taskWidgets = ModuleLoader.loadModuleClasses(
            moduleDirectoryPath=moduleDirectoryPath,
            moduleBaseClass=TaskWidget,
            fileNameEndsWith='taskwidget.py',
        )
        for taskName in taskWidgets.keys():
            LOGGER.info(f'TaskWidgetManager: Loaded task type {taskName}')
        return taskWidgets


from typing import Any

from PySide6.QtWidgets import QWidget, QLabel, QSizePolicy
from PySide6.QtGui import QFont

from mosamaticdesktop.tasks.parameter import Parameter


class LabelParameter(Parameter):
    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, parent: QWidget=None) -> None:
        super(LabelParameter, self).__init__(
            name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, parent=parent)
        self.initUi()

    def initUi(self) -> None:
        font = QFont()
        font.setItalic(True)
        label = QLabel('')
        label.setFont(font)
        label.setWordWrap(True)
        label.setText(self.labelText())
        policy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        label.setSizePolicy(policy)
        self.layout().addWidget(label)

    def copy(self):
        return LabelParameter(
            name=self.name(), 
            labelText=self.labelText(), 
            optional=self.optional(), 
            visible=self.visible(), 
            defaultValue=self.defaultValue(),
        )


from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.calculatebodycompositionmetricstask.calculatebodycompositionmetricstask import CalculateBodyCompositionMetricsTaskTask


class CalculateBodyCompositionMetricsTaskWidget(TaskWidget):
    def __init__(self, progressBar: QProgressBar) -> None:
        super(CalculateBodyCompositionMetricsTaskWidget, self).__init__(taskType=CalculateBodyCompositionMetricsTaskTask, progressBar=progressBar)
        self.addDescriptionParameter(
            name='description',
            description='Calculates body composition metrics on predicted segmenations and TAG files if available'
        )
        self.addFileSetParameter(
            name='inputFileSetName',
            labelText='Input DICOM (and TAG) File Set',
        )
        self.addFileSetParameter(
            name='inputSegmentationFileSetName',
            labelText='Input Segmentation File Set',
        )
        self.addPathParameter(
            name='patientHeightsCsvFilePath',
            labelText='Patient Height CSV File Path (File Name, Height (m))',
            optional=True,
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )
    
    def validate(self) -> None:
        pass


from typing import Any

from PySide6.QtWidgets import QWidget, QVBoxLayout

class Parameter(QWidget):
    @classmethod
    def NAME(cls):
        return cls.__qualname__

    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, parent: QWidget=None) -> None:
        super(Parameter, self).__init__(parent=parent)
        self._name = name
        self._labelText = labelText
        if not optional and not self._labelText.endswith('*'):
            self._labelText = labelText + '*'
        self._optional = optional
        self._visible = visible
        self._defaultValue = defaultValue
        self._value = None
        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(5, 5, 5, 5)
        self._layout.setSpacing(5)
        self.setLayout(self._layout)

    def name(self) -> str:
        return self._name

    def labelText(self) -> str:
        return self._labelText
    
    def optional(self) -> bool:
        return self._optional
    
    def visible(self) -> bool:
        return self._visible
    
    def defaultValue(self) -> Any:
        return self._defaultValue
    
    def value(self) -> Any:
        return self._value
    
    def setValue(self, value: Any) -> None:
        self._value = value

    def layout(self) -> QVBoxLayout:
        return self._layout
    
    def copy(self):
        raise NotImplementedError()


import inspect

from typing import Any, List

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QProgressBar, QProgressBar, QDialog, QMessageBox

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.tasks.taskwidgetexception import TaskWidgetException
from mosamaticdesktop.tasks.taskwidgetparameterdialog import TaskWidgetParameterDialog
from mosamaticdesktop.tasks.taskruninfodialog import TaskRunInfoDialog
from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parametercopier import ParameterCopier
from mosamaticdesktop.tasks.descriptionparameter import DescriptionParameter
from mosamaticdesktop.tasks.labelparameter import LabelParameter
from mosamaticdesktop.tasks.filesetparameter import FileSetParameter
from mosamaticdesktop.tasks.multifilesetparameter import MultiFileSetParameter
from mosamaticdesktop.tasks.pathparameter import PathParameter
from mosamaticdesktop.tasks.textparameter import TextParameter
from mosamaticdesktop.tasks.integerparameter import IntegerParameter
from mosamaticdesktop.tasks.floatingpointparameter import FloatingPointParameter
from mosamaticdesktop.tasks.booleanparameter import BooleanParameter
from mosamaticdesktop.tasks.optiongroupparameter import OptionGroupParameter
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class TaskWidget(QWidget):    
    # Trick to return the child class name when NAME is retrieved
    # https://chat.openai.com/c/b7bd6334-5ec3-40e3-9af1-93405c68d795
    @classmethod
    def NAME(cls):
        # Returns class name of child classes (and strips of last 6 characters to get the task name itself)
        return cls.__qualname__[:-6]
    
    def __init__(self, taskType: Task, progressBar: QProgressBar) -> None:
        super(TaskWidget, self).__init__()
        if not inspect.isclass(taskType):
            raise TaskWidgetException('TaskWidget: argument taskType should be a class')
        self._taskType = taskType        
        self._progressBar = progressBar
        self._task = None
        self._taskParameters = {}
        # We need this class to copy parameters to the TaskWidgetParameterDialog, otherwise the
        # parameters get deleted by C++ after the dialog closes.
        self._taskParameterCopier = ParameterCopier()
        ######
        # self._progressBar = None
        # self._progressBarLabel = None
        self._startButton = None
        self._cancelButton = None
        self._settingsButton = None
        self._runInfoButton = None
        self._test = False
        self.initUi()

    def name(self) -> str:
        return self.__class__.__name__
    
    def setTest(self, test: bool) -> None:
        self._test = test
    
    def initUi(self) -> None:
        # self._progressBarLabel = QLabel('0 %')
        # labelLayout = QHBoxLayout()
        # labelLayout.addItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        # labelLayout.addWidget(self._progressBarLabel)
        # labelLayout.addItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        # self._progressBar = QProgressBar(self)
        # self._progressBar.setRange(0, 100)
        # self._progressBar.setValue(0)
        self._startButton = QPushButton('Start')
        self._startButton.setObjectName('startButton') # for testing
        self._startButton.clicked.connect(self.startTask)
        self._cancelButton = QPushButton('Cancel')
        self._cancelButton.setObjectName('cancelButton') # for testing
        self._cancelButton.setEnabled(False)
        self._cancelButton.clicked.connect(self.cancelTask)
        self._settingsButton = QPushButton('Set Task Parameters...')
        self._settingsButton.setObjectName('settingsButton') # for testing
        self._settingsButton.clicked.connect(self.showTaskWidgetParameterDialog)
        self._runInfoButton = QPushButton('Run Info...')
        self._runInfoButton.setObjectName('runInfoButton')
        self._runInfoButton.setEnabled(False)
        self._runInfoButton.clicked.connect(self.showTaskRunInfoDialog)
        buttonLayout = QHBoxLayout()
        buttonLayout.addWidget(self._startButton)
        buttonLayout.addWidget(self._cancelButton)
        buttonLayout.addWidget(self._settingsButton)
        buttonLayout.addWidget(self._runInfoButton)
        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignTop)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setSpacing(5)
        # if OperatingSystem.isDarwin():
        #     layout.addLayout(labelLayout)
        # layout.addWidget(self._progressBar)
        layout.addLayout(buttonLayout)
        self.setLayout(layout)

    # Task parameters
        
    def taskParameter(self, name: str) -> Parameter:
        if name in self._taskParameters.keys():
            return self._taskParameters[name]
        return None
        
    def addDescriptionParameter(self, name: str, description: str) -> Parameter:
        parameter = DescriptionParameter(name=name, description=description)
        self._taskParameters[parameter.name()] = parameter
        return parameter

    def addLabelParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = LabelParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._taskParameters[parameter.name()] = parameter
        return parameter

    def addFileSetParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = FileSetParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._taskParameters[parameter.name()] = parameter
        return parameter
    
    def addMultiFileSetParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = MultiFileSetParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._taskParameters[parameter.name()] = parameter
        return parameter

    def addPathParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = PathParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._taskParameters[parameter.name()] = parameter
        return parameter

    def addTextParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = TextParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._taskParameters[parameter.name()] = parameter
        return parameter

    def addIntegerParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, minimum: int=0, maximum: int=100, step: int=1) -> Parameter:
        parameter = IntegerParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, minimum=minimum, maximum=maximum, step=step)
        LOGGER.info(f'TaskWidget.addIntegerParameter() value={parameter.name()}')
        self._taskParameters[parameter.name()] = parameter
        return parameter

    def addFloatingPointParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = FloatingPointParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._taskParameters[parameter.name()] = parameter
        return parameter
    
    def addBooleanParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = BooleanParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._taskParameters[parameter.name()] = parameter
        return parameter

    def addOptionGroupParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, options: List[str]=[]) -> Parameter:
        parameter = OptionGroupParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, options=options)
        self._taskParameters[parameter.name()] = parameter
        return parameter

    # Task execution

    def startTask(self) -> None:
        self._task = self._taskType() # instantiate class
        self._task.setParameters(parameters=self._taskParameters)
        self._task.signal().progress.connect(self.taskProgress)
        self._task.signal().finished.connect(self.taskFinished)
        self._task.start()
        self._startButton.setEnabled(False)
        self._cancelButton.setEnabled(True)
        self._runInfoButton.setEnabled(False)
        self._progressBar.setValue(0)

    def cancelTask(self) -> None:
        if self._task:
            self._task.cancel()
            if not self._test:
                self._runInfoButton.setEnabled(True)
                self._cancelButton.setEnabled(False)
                self._startButton.setEnabled(True)
                self._progressBar.setValue(0)

    def showTaskWidgetParameterDialog(self) -> None:
        # Hack: we need to explicitly copy each parameter before passing them to
        # the TaskWidgetParameterDialog. If not, the Qt C++ backend will delete
        # the parameters if we try to show the dialog twice.
        parameters = {}
        for parameter in self._taskParameters.values():
            parameters[parameter.name()] = self._taskParameterCopier.makeCopy(parameter)
        # Show task widget parameter dialog
        dialog = TaskWidgetParameterDialog(title=self.name(), parameters=parameters)
        result = dialog.show()
        if result == QDialog.Accepted:
            self._taskParameters = dialog.parameters()
            self.validate()

    def showTaskRunInfoDialog(self) -> None:
        if self._task:
            dialog = TaskRunInfoDialog(
                taskName=self._task.name(), errors=self._task.errors(), warnings=self._task.warnings(), info=self._task.info())
            dialog.show()
        
    def taskProgress(self, progress: int) -> None:
        if not self._test:
            self._progressBar.setValue(progress)

    def taskFinished(self, value: bool) -> None:
        if not self._test:
            self._runInfoButton.setEnabled(True)
            self._cancelButton.setEnabled(False)
            self._startButton.setEnabled(True)
            if self._task and (self._task.hasErrors() or self._task.hasWarnings()):
                self.showTaskRunInfoDialog()

    # Validation
        
    def validate(self) -> None:
        raise NotImplementedError()
    
    def showValidationError(self, parameterName: str, message: str) -> None:
        QMessageBox.critical(self, f'Validation Error for parameter "{parameterName}"', message)

    # Status (only used for testing)
        
    def taskIsIdle(self) -> bool:
        if self._task:
            return self._task.statusIsIdle()
        raise TaskWidgetException('TaskWidget: cannot determine task idle status, start it first')
        
    def taskIsStart(self) -> bool:
        if self._task:
            return self._task.statusIsStart()
        raise TaskWidgetException('TaskWidget: cannot determine task start status, start it first')
        
    def taskIsRunning(self) -> bool:
        if self._task:
            return self._task.statusIsRunning()
        raise TaskWidgetException('TaskWidget: cannot determine task running status, start it first')
        
    def taskIsCanceling(self) -> bool:
        if self._task:
            return self._task.statusIsCanceling()
        raise TaskWidgetException('TaskWidget: cannot determine task canceled status, start it first')
        
    def taskIsCanceled(self) -> bool:
        if self._task:
            return self._task.statusIsCanceled()
        raise TaskWidgetException('TaskWidget: cannot determine task canceled status, start it first')
        
    def taskIsFinished(self) -> bool:
        if self._task:
            return self._task.statusIsFinished()
        raise TaskWidgetException('TaskWidget: cannot determine task finished status, start it first')
    
    def taskIsError(self) -> bool:
        if self._task:
            return self._task.statusIsError()
        raise TaskWidgetException('TaskWidget: cannot determine task error status, start it first')


from typing import Any

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QVBoxLayout, QCheckBox

from mosamaticdesktop.tasks.parameter import Parameter


class BooleanParameter(Parameter):
    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, parent: QWidget=None) -> None:
        super(BooleanParameter, self).__init__(
            name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, parent=parent)
        self._booleanCheckBox = None
        self.initUi()

    def initUi(self) -> None:
        self._booleanCheckBox = QCheckBox(self)
        self._booleanCheckBox.setText(self.labelText())
        if self.defaultValue() is not None:
            self._booleanCheckBox.setCheckState(Qt.Checked if self.defaultValue() else Qt.Unchecked)
            self.stateChanged(state=self._booleanCheckBox.checkState())
        self._booleanCheckBox.stateChanged.connect(self.stateChanged)
        self.layout().addWidget(self._booleanCheckBox)

    def stateChanged(self, state: int) -> None:
        self.setValue(True if state == Qt.Checked else False)

    def copy(self):
        return BooleanParameter(
            name=self.name(), 
            labelText=self.labelText(), 
            optional=self.optional(),             
            visible=self.visible(), 
            defaultValue=self.defaultValue()
        )